@@ -39,6 +39,7 @@
 
 #include "qtiffhandler_p.h"
 #include <qvariant.h>
+#include <qcolorspace.h>
 #include <qdebug.h>
 #include <qimage.h>
 #include <qglobal.h>
@@ -204,9 +205,14 @@ bool QTiffHandlerPrivate::canRead(QIODevice *device)
 
     // current implementation uses TIFFClientOpen which needs to be
     // able to seek, so sequential devices are not supported
-    QByteArray header = device->peek(4);
-    return header == QByteArray::fromRawData("\x49\x49\x2A\x00", 4)
-           || header == QByteArray::fromRawData("\x4D\x4D\x00\x2A", 4);
+    char h[4];
+    if (device->peek(h, 4) != 4)
+        return false;
+    if ((h[0] == 0x49 && h[1] == 0x49) && (h[2] == 0x2a || h[2] == 0x2b) && h[3] == 0)
+        return true; // Little endian, classic or bigtiff
+    if ((h[0] == 0x4d && h[1] == 0x4d) && h[2] == 0 && (h[3] == 0x2a || h[3] == 0x2b))
+        return true; // Big endian, classic or bigtiff
+    return false;
 }
 
 bool QTiffHandlerPrivate::openForRead(QIODevice *device)
@@ -272,6 +278,8 @@ bool QTiffHandlerPrivate::readHeaders(QIODevice *device)
         format = QImage::Format_Mono;
     else if (photometric == PHOTOMETRIC_MINISBLACK && bitPerSample == 8 && samplesPerPixel == 1)
         format = QImage::Format_Grayscale8;
+    else if (photometric == PHOTOMETRIC_MINISBLACK && bitPerSample == 16 && samplesPerPixel == 1)
+        format = QImage::Format_Grayscale16;
     else if ((grayscale || photometric == PHOTOMETRIC_PALETTE) && bitPerSample == 8 && samplesPerPixel == 1)
         format = QImage::Format_Indexed8;
     else if (samplesPerPixel < 4)
@@ -391,9 +399,10 @@ bool QTiffHandler::read(QImage *image)
                 }
 
                 for (int i = 0; i<tableSize ;++i) {
-                    const int red = redTable[i] / 257;
-                    const int green = greenTable[i] / 257;
-                    const int blue = blueTable[i] / 257;
+                    // emulate libtiff behavior for 16->8 bit color map conversion: just ignore the lower 8 bits
+                    const int red = redTable[i] >> 8;
+                    const int green = greenTable[i] >> 8;
+                    const int blue = blueTable[i] >> 8;
                     qtColorTable[i] = qRgb(red, green, blue);
                 }
             }
@@ -402,9 +411,11 @@ bool QTiffHandler::read(QImage *image)
         }
     }
     bool format8bit = (format == QImage::Format_Mono || format == QImage::Format_Indexed8 || format == QImage::Format_Grayscale8);
+    bool format16bit = (format == QImage::Format_Grayscale16);
     bool format64bit = (format == QImage::Format_RGBX64 || format == QImage::Format_RGBA64 || format == QImage::Format_RGBA64_Premultiplied);
 
-    if (format8bit || format64bit) {
+    // Formats we read directly, instead of over RGBA32:
+    if (format8bit || format16bit || format64bit) {
         int bytesPerPixel = image->depth() / 8;
         if (format == QImage::Format_RGBX64)
             bytesPerPixel = 6;
@@ -483,6 +494,15 @@ bool QTiffHandler::read(QImage *image)
         }
     }
 
+    uint32 count;
+    void *profile;
+    if (TIFFGetField(tiff, TIFFTAG_ICCPROFILE, &count, &profile)) {
+        QByteArray iccProfile(reinterpret_cast<const char *>(profile), count);
+        image->setColorSpace(QColorSpace::fromIccProfile(iccProfile));
+    }
+    // We do not handle colorimetric metadat not on ICC profile form, it seems to be a lot
+    // less common, and would need additional API in QColorSpace.
+
     return true;
 }
 
@@ -513,6 +533,7 @@ static QVector<QRgb> effectiveColorTable(const QImage &image)
             colors[i] = qRgba(0, 0, 0, i);
         break;
     case QImage::Format_Grayscale8:
+    case QImage::Format_Grayscale16:
         colors.resize(256);
         for (int i = 0; i < 256; ++i)
             colors[i] = qRgb(i, i, i);
@@ -586,7 +607,14 @@ bool QTiffHandler::write(const QImage &image)
         TIFFClose(tiff);
         return false;
     }
-
+    // set color space
+    if (image.colorSpace().isValid()) {
+        QByteArray iccProfile = image.colorSpace().iccProfile();
+        if (!TIFFSetField(tiff, TIFFTAG_ICCPROFILE, iccProfile.size(), reinterpret_cast<const void *>(iccProfile.constData()))) {
+            TIFFClose(tiff);
+            return false;
+        }
+    }
     // configure image depth
     const QImage::Format format = image.format();
     if (format == QImage::Format_Mono || format == QImage::Format_MonoLSB) {
@@ -602,8 +630,8 @@ bool QTiffHandler::write(const QImage &image)
         }
 
         // try to do the conversion in chunks no greater than 16 MB
-        int chunks = (width * height / (1024 * 1024 * 16)) + 1;
-        int chunkHeight = qMax(height / chunks, 1);
+        const int chunks = int(image.sizeInBytes() / (1024 * 1024 * 16)) + 1;
+        const int chunkHeight = qMax(height / chunks, 1);
 
         int y = 0;
         while (y < height) {
@@ -622,6 +650,7 @@ bool QTiffHandler::write(const QImage &image)
         TIFFClose(tiff);
     } else if (format == QImage::Format_Indexed8
                || format == QImage::Format_Grayscale8
+               || format == QImage::Format_Grayscale16
                || format == QImage::Format_Alpha8) {
         QVector<QRgb> colorTable = effectiveColorTable(image);
         bool isGrayscale = checkGrayscale(colorTable);
@@ -631,7 +660,7 @@ bool QTiffHandler::write(const QImage &image)
                 photometric = PHOTOMETRIC_MINISWHITE;
             if (!TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, photometric)
                     || !TIFFSetField(tiff, TIFFTAG_COMPRESSION, compression == NoCompression ? COMPRESSION_NONE : COMPRESSION_LZW)
-                    || !TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, 8)
+                    || !TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, image.depth())
                     || !TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, defaultStripSize(tiff))) {
                 TIFFClose(tiff);
                 return false;
@@ -669,22 +698,10 @@ bool QTiffHandler::write(const QImage &image)
         }
 
         //// write the data
-        // try to do the conversion in chunks no greater than 16 MB
-        int chunks = (width * height/ (1024 * 1024 * 16)) + 1;
-        int chunkHeight = qMax(height / chunks, 1);
-
-        int y = 0;
-        while (y < height) {
-            QImage chunk = image.copy(0, y, width, qMin(chunkHeight, height - y));
-
-            int chunkStart = y;
-            int chunkEnd = y + chunk.height();
-            while (y < chunkEnd) {
-                if (TIFFWriteScanline(tiff, reinterpret_cast<uint32 *>(chunk.scanLine(y - chunkStart)), y) != 1) {
-                    TIFFClose(tiff);
-                    return false;
-                }
-                ++y;
+        for (int y = 0; y < height; ++y) {
+            if (TIFFWriteScanline(tiff, const_cast<uchar *>(image.scanLine(y)), y) != 1) {
+                TIFFClose(tiff);
+                return false;
             }
         }
         TIFFClose(tiff);
@@ -742,7 +759,7 @@ bool QTiffHandler::write(const QImage &image)
             return false;
         }
         // try to do the RGB888 conversion in chunks no greater than 16 MB
-        const int chunks = (width * height * 3 / (1024 * 1024 * 16)) + 1;
+        const int chunks = int(image.sizeInBytes() / (1024 * 1024 * 16)) + 1;
         const int chunkHeight = qMax(height / chunks, 1);
 
         int y = 0;
@@ -774,7 +791,7 @@ bool QTiffHandler::write(const QImage &image)
             return false;
         }
         // try to do the RGBA8888 conversion in chunks no greater than 16 MB
-        const int chunks = (width * height * 4 / (1024 * 1024 * 16)) + 1;
+        const int chunks = int(image.sizeInBytes() / (1024 * 1024 * 16)) + 1;
         const int chunkHeight = qMax(height / chunks, 1);
 
         const QImage::Format format = premultiplied ? QImage::Format_RGBA8888_Premultiplied
@@ -799,11 +816,6 @@ bool QTiffHandler::write(const QImage &image)
     return true;
 }
 
-QByteArray QTiffHandler::name() const
-{
-    return "tiff";
-}
-
 QVariant QTiffHandler::option(ImageOption option) const
 {
     if (option == Size && canRead()) {
